import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * PS4
 * Authors: Logan Tucker and Tucker Simpson
 */

public class Bacon {

    // create global maps to hold information from the data files. this will be useful when placing information into the
    // graph. Also store a path to the data's location in the users device
    TreeMap<Integer, String> iDToActor;
    String actorsPathName;

    TreeMap<Integer, String> iDToMovie;
    String moviesPathName;

    TreeMap<Integer, ArrayList<Integer>> movieIDToActorID;
    String movieActorPathName;

    AdjacencyMapGraph<String, TreeSet<String>> actorGraph;


    public Bacon (String actorsPathName, String moviesPathName, String movieActorsPathName) {
        this.iDToActor = new TreeMap<>();
        this.actorsPathName = actorsPathName;

        this.iDToMovie = new TreeMap<>();
        this.moviesPathName = moviesPathName;

        this.movieIDToActorID = new TreeMap<>();
        this.movieActorPathName = movieActorsPathName;
    }


    /**
     * This method makes a map from a data file holding "|" separated integers (Integer|Integer).
     * @param pathName
     * @return a map with key: Movie ID and value: Actor IDs of stars
     */
    public TreeMap<Integer, ArrayList<Integer>> actorMovieFileToMap(String pathName) {

        TreeMap<Integer, ArrayList<Integer>> actorMovieMap = new TreeMap<Integer, ArrayList<Integer>>();

        BufferedReader input;

        // use try catch to intercept bogus files and paths
        try {
            input = new BufferedReader(new FileReader(pathName));

            // read first line of file and store it as a string
            String line = input.readLine();
            while (line != null) {

                // separate the key and value from within the line
                String[] keyAndValue = line.split("\\|");
                int actorID = Integer.parseInt(keyAndValue[0]);
                int movieID = Integer.parseInt(keyAndValue[1]);

                // if the map does not already contain this key
                if (!actorMovieMap.containsKey(actorID)) {

                    // add the key and value (new array list) to the map
                    ArrayList<Integer> actorIds = new ArrayList<>();
                    actorIds.add(movieID);
                    actorMovieMap.put(actorID, actorIds);
                }

                // if the map already contains this key (movie)
                else {

                    // add this new actor to its cast list (value)
                    actorMovieMap.get(actorID).add(movieID);
                }

                // read the next line
                line = input.readLine();
            }
            input.close();
        }
        catch(Exception FileNotFoundException) {
            System.out.println("File " + pathName + " not found");
        }
        return actorMovieMap;
    }


    /**
     * This method makes a map from a data file holding "|" separated Integer and Strings (Integer|String).
     * @param pathName
     * @return a map with key: Movie or Actor ID and value: name
     */
    public TreeMap<Integer, String> fileToMap(String pathName) {

        TreeMap<Integer, String> map = new TreeMap<>();

        BufferedReader input;

        // use try catch to intercept bogus files and paths
        try {
            input = new BufferedReader(new FileReader(pathName));

            // read first line of file and store it as a string
            String line = input.readLine();
            while (line != null) {

                // separate the key and value from within the line
                String[] keyAndValue = line.split("\\|");

                // add to the map
                map.put(Integer.parseInt(keyAndValue[0]), keyAndValue[1]);

                // read the next line
                line = input.readLine();
            }
            input.close();
        }
        catch(Exception FileNotFoundException) {
            System.out.println("File " + pathName + " not found");
        }
        return map;
    }


    /**
     * This method calls the appropriate file to map converters and stores maps globally
     */
    public void getMaps() {
        iDToActor = fileToMap(actorsPathName);
        iDToMovie = fileToMap(moviesPathName);
        movieIDToActorID = actorMovieFileToMap(movieActorPathName);
    }


    /**
     * This method pulls data from maps to create a graph
     * @return a graph of actors connected to costar by movie title
     */
    public AdjacencyMapGraph<String, TreeSet<String>> getGraph() {
        // create maps of data
        getMaps();

        // create a new graph
        actorGraph = new AdjacencyMapGraph<>();

        // create a vertex for each actor
        for (Integer a : iDToActor.keySet()) {
            actorGraph.insertVertex(iDToActor.get(a));
        }

        // for each movie
        for (Integer m : movieIDToActorID.keySet()) {

            // for each actor i in current movie
            for (int i = 0; i < movieIDToActorID.get(m).size(); i++) {

                // loop through every actor j and connect them
                for (int j = 0; j < movieIDToActorID.get(m).size(); j++) {

                    // get the names of each actor and the movie from their ID
                    String v1 = iDToActor.get(movieIDToActorID.get(m).get(i));
                    String v2 = iDToActor.get(movieIDToActorID.get(m).get(j));
                    String movie = iDToMovie.get(m);

                    // if these actors are different humans (costars)
                    if (j != i ){

                        // if this relationship is being discovered (they have no previous shared movies listed)
                        if(!actorGraph.hasEdge(v1, v2)) {

                            // create a connection between the two stars
                            TreeSet<String> moviesShared = new TreeSet<>();
                            moviesShared.add(movie);
                            actorGraph.insertUndirected(v1, v2, moviesShared);
                        }

                        // if they have already been noted to have co-stared together in a DIFFERENT movie
                        else {

                            // add this movie to their list of shared films (held in the label of their connection)
                            actorGraph.getLabel(v1, v2).add(movie);
                        }
                    }
                }
            }
        }
        return actorGraph;
    }

    public static void main(String[] args) {
        Bacon tester = new Bacon("/Users/tuckersimpson/IdeaProjects/cs10/ps4_inputs/actors.txt", "/Users/tuckersimpson/IdeaProjects/cs10/ps4_inputs/movies.txt", "/Users/tuckersimpson/IdeaProjects/cs10/ps4_inputs/movie-actors.txt");
        Graph<String, TreeSet<String>> testGraph = tester.getGraph();
        Graph <String, TreeSet<String>> fidelBFS = BaconGraphLib.bfs(testGraph, "Nita Naldi");
        System.out.println(fidelBFS);
        System.out.println(BaconGraphLib.averageSeparation(fidelBFS, "Nita Naldi"));
        Graph <String, TreeSet<String>>  julieBFS = BaconGraphLib.bfs(testGraph, "Julie Bishop");
        System.out.println(BaconGraphLib.averageSeparation(julieBFS, "Julie Bishop"));


    }
}
