import java.util.*;

/**
 * PS4
 * Authors: Logan Tucker and Tucker Simpson
 */

public class BaconGameInterface {

    String currCenter;

    Bacon currBacon;

    boolean gameRunning;

    Graph<String, TreeSet<String>> baconGraph;
    Graph<String, TreeSet<String>> ShortestPathTree;


    public BaconGameInterface(String actorsPathName, String moviesPathName, String movieActorPathName){

        Bacon currBacon = new Bacon(actorsPathName, moviesPathName, movieActorPathName);

        currCenter = "Kevin Bacon";

        baconGraph = currBacon.getGraph();

        ShortestPathTree  = BaconGraphLib.bfs(baconGraph, currCenter);

        gameRunning = true;
    }


    /**
     * create the shortest path tree given a center vertex
     * @param newCenter
     */
    public void makeCenter(String newCenter){

        ShortestPathTree = BaconGraphLib.bfs(baconGraph, newCenter);
        currCenter = newCenter;
    }


    /**
     * find the shortest path from a given vertex to the center
     * @param v
     * @return the shortest path from a given vertex to the center
     */
    public List<String> shortestPath(String v){
        return BaconGraphLib.getPath(ShortestPathTree, v);
    }


    /**
     * this method maps vertices to their degree of separation to the center
     * @return map of key: degree of separation from graph center and value: actors
     */
    public Map<Integer, Set<String>> vertsBySep(){

        // create storage
        Map<Integer, Set<String>> vertsMap = new HashMap();

        // add the current center to the map
        Set<String> actors = new HashSet<>();
        actors.add(currCenter);
        vertsMap.put(0, actors);

        // note the lowest degree of separation for non centers to be 1
        int currSep = 1;

        // add all other actors to the map
        vertsMap = vertsBySepHelper(vertsMap, currSep, currCenter);

        return vertsMap;
    }


    /**
     * add all actors to map with key: degree of separation from graph center
     * @param vertsMap
     * @param currSep
     * @param currActor
     * @return map of key: degree of separation from graph center and value: actors
     */
    public Map<Integer, Set<String>> vertsBySepHelper(Map<Integer, Set<String>> vertsMap, int currSep, String currActor){

        // for each neighbor of the current actor
        for(String neighbor : ShortestPathTree.inNeighbors(currActor)){

            // if this is the first actor found with current degree of separation
            if (!vertsMap.containsKey(currSep)) {

                // add a new key to the map
                Set<String> actorList = new HashSet<>();
                actorList.add(neighbor);
                vertsMap.put(currSep, actorList);
            }

            // if the map already has actors at the current degree of separation
            else {
                vertsMap.get(currSep).add(neighbor);
            }

            // call vertsBySepHelper recursively with an incremented separation and current neighbor
            vertsBySepHelper(vertsMap, currSep+1, neighbor);
        }
        return vertsMap;
    }


    /**
     * @return the number of actors in the shortest path tree
     */
    public int numActorsConnected(){
        return ShortestPathTree.numVertices()-1;
    }


    /**
     * @param desiredCenter
     * @return average path length (degree of separation) from vertex's to center of graph
     */
    public double avgPathLength(String desiredCenter){

        if(!currCenter.equals(desiredCenter)) {
            makeCenter(desiredCenter);
        }
        return BaconGraphLib.averageSeparation(ShortestPathTree, currCenter);
    }


    /**
     * @return a list of actors in order of in degree
     */
    public List<String> getActorsByDegree(){
        return BaconGraphLib.verticesByInDegree(baconGraph);
    }


    /**
     *
     * @return a list of actors sorted in order of average separation when they are the center of the universe
     */
    public List<String> getActorsByAvgSep(int num){

        // create storage
        PriorityQueue<SepAndActor> actorQueue;
        ArrayList<String> sortedActorList = new ArrayList<>();

        // do not run if asked to return an empty list
        if (num != 0) {

            // create a min or max priority queue of vertices, sorted by avg separation when a vertex is a center
            if (num < 0) {
                actorQueue = new PriorityQueue<SepAndActor>((SepAndActor e1, SepAndActor e2) -> Double.compare(e2.avgSep, e1.avgSep));
            }
            else {
                actorQueue = new PriorityQueue<SepAndActor>((SepAndActor e1, SepAndActor e2) -> Double.compare(e1.avgSep, e2.avgSep));
            }

            // for every actor in the graph
            for(String actor : baconGraph.vertices()) {

                // get its average degree of separation when the center
                double avg = BaconGraphLib.averageSeparation(BaconGraphLib.bfs(baconGraph, actor), actor);

                // store the actors information and add it to the queue
                SepAndActor sna = new SepAndActor(avg, actor);
                actorQueue.add(sna);
            }

            // add users desired number of actors in order to an array list
            for (int i = 0; i < Math.abs(num); i++) {
                sortedActorList.add(actorQueue.remove().actor);
            }

        }
        return sortedActorList;
    }


    /**
     * print the directions of the game
     */
    public void newGame() {

        // begin the game with kevin bacon as the graph center
        currCenter = "Kevin Bacon";



        String head = "Commands: \nc <#>: list top (positive number) or bottom (negative) <#> centers of the " +
                "universe, sorted by average separation \nd <low> <high>: list actors sorted by degree (high to low)" +
                ", with degree between low and high \ni: list actors with infinite separation from the current center \n" +
                "p <name>: find path from <name> to current center of the universe \ns <low> <high>: list actors sorted " +
                "by non-infinite separation from the current center, with separation between low and high \nu <name>: " +
                "make <name> the center of the universe \nq: quit game \n \n"+ currCenter + " is now the center of the " +
                "acting universe, connected to " + (ShortestPathTree.numVertices()-1) + "/" + (baconGraph.numVertices()-1) +
                " actors with average separation " + avgPathLength(currCenter) + '\n';

        // mark the game as running
        gameRunning = true;

        System.out.println(head);
    }

    /**
     * Gets user input when looking for int, accounts for edge cases
     * @param inputPrompt String to prompt user input
     * @param onlyPos true if user input needs to be a positive number
     * @return return the user's input as an int
     */
    public int getInt(String inputPrompt, boolean onlyPos){
        // Prompt and get user input
        Scanner myScanner = new Scanner(System.in);
        System.out.print(inputPrompt);
        String stringNum = myScanner.nextLine();

        int num;

        // Try to parse int, catch exception
        try{
            num = Integer.parseInt(stringNum);

            // make sure the int is positive if it needs to be positive
            if(!onlyPos) return num;
            else if (num<0){
                System.out.println("Invalid input. Enter a positive number.");
                return getInt(inputPrompt, onlyPos);
            }
        }
        // catch input with invalid format
        catch(Exception NumberFormatException){
            System.out.println("Invalid input, try again.");
            return getInt(inputPrompt, onlyPos);
        }
        return num;
    }

    /**
     * Prompt user to input an actor, keep prompting until they give a name of an actor in the current universe
     * @param inputPrompt prompt for user input
     * @return actor name
     */
    public String getActor(String inputPrompt){
        // Prompt user input
        Scanner myScanner = new Scanner(System.in);
        System.out.print(inputPrompt);
        String actor = myScanner.nextLine();

        // Make sure actor is in this universe
        if(!baconGraph.hasVertex(actor)){
            return getActor("Actor is not in universe, enter a new actor: ");
        }else{
            return actor;
        }

    }

    /**
     * get user input for command, check to make sure input is a valid command
     * @return user's input as a string
     */
    public String getCommand(){
        Scanner myScanner = new Scanner(System.in);
        System.out.print("Enter letter command: ");

        // read user input
        String command = myScanner.nextLine();
        if (command.equals("c")||command.equals("d")||command.equals("i")||command.equals("p")||command.equals("s")
                ||command.equals("u")||command.equals("q")){
            return command;
        }else{
            System.out.println("Command unknown\n");
            return getCommand();
        }
    }

    /**
     * quit the game, return all global values to original state
     */
    public void quitGame(){
        System.out.println("Ending game.");
        gameRunning = false;
        currCenter = null;
        baconGraph = null;
        ShortestPathTree = null;
    }


    /**
     * interact with the user
     */
    public void handleInput() {

        // get command
        String command = getCommand();

        // make new center of the universe
        if (command.equals("u")) {

            String actor = getActor("Enter actor name: ");

            makeCenter(actor);
            System.out.println(actor + " is now the center of the acting universe, connected to " +
                    (ShortestPathTree.numVertices() - 1) + "/" + (baconGraph.numVertices() - 1) +
                    " actors with average separation " + avgPathLength(actor) + '\n');
            System.out.println(actor + " game >");

        }

        // Quit game
        else if (command.equals("q")) {
            quitGame();
        }

        // list actors with infinite separation from the current center
        else if (command.equals("i")) {

            // get actors with infinite degree of separation from center
            Set<String> missingActors = BaconGraphLib.missingVertices(baconGraph, ShortestPathTree);
            System.out.println("The following actors are infinitely separated from the current center: \n" +
                    missingActors + "\n");
        }

        // get path from actor to the center of the universe
        else if (command.equals("p")) {

            // get actor name
            String actor = getActor("Enter actor name: ");
            List<String> path;

            if (ShortestPathTree.hasVertex(actor)) {

                // get the shortest path from actor to center
                path = shortestPath(actor);

                // return information
                System.out.println(actor + "'s " + currCenter + " number is " + (path.size() - 1));

                // print each connection on the path
                for (int i = 1; i < path.size(); i++) {
                    System.out.println(actor + " appeared in " + ShortestPathTree.getLabel(actor, path.get(i)) +
                            " with " + path.get(i));

                    actor = path.get(i);
                }
                System.out.println();
            } else {
                System.out.println(actor + "'s " + currCenter + " number is infinite.");
            }
        }

        // list top (positive number) or bottom (negative) <#> centers of the universe, sorted by average separation
        else if (command.equals("c")) {

            // get number of actors to print
            int num = getInt("Enter number of actors to print: ", false);

            // get list of actors sorted by avg separation when center
            List<String> actorsBySep = getActorsByAvgSep(num);

            if (num < 0) {
                System.out.println("The " + Math.abs(num) + " actor(s) with the highest average separation are: " +
                        actorsBySep);
            } else {
                System.out.println("The " + num + " actor(s) with the lowest average separation are: " + actorsBySep);
            }

            System.out.println();
        }

        // list actors sorted by degree, with degree between low and high
        else if(command.equals("d")) {

            // get upper and lower bounds
            int lowNum = getInt("Enter lower bound: ", true);
            int upNum = getInt("Enter upper bound: ", true);

            // get list of vertices sorted by in degree
            List<String> vertsByInDegree = BaconGraphLib.verticesByInDegree(baconGraph);
            List<String> returnList = new ArrayList<>();

            for (String vert : vertsByInDegree) {

                // once you have filled the bounds stop inspecting vertices
                if (baconGraph.inDegree(vert) < lowNum) break;
                // if the vertex's in degree is within the bounds
                else if (baconGraph.inDegree(vert) <= upNum) {
                    returnList.add(vert);
                }

            }

            System.out.println("The vertices with in degrees between " + lowNum + " and " + upNum + " are: " +
                    returnList + "\n");
        }

        // list actors sorted by non-infinite separation from the current center, with separation between low and high
        else if(command.equals("s")) {

            // prompt user for input
            int lowNum = getInt("Enter lower bound: ", true);
            int upNum = getInt("Enter upper bound: ", true);

            // get map of vertices to degrees of separation from center
            Map<Integer, Set<String>> degreeMap = vertsBySep();
            ArrayList<String> actorList = new ArrayList<>();

            // add actors within bounds to a list
            for (int i = lowNum; i <= upNum; i++) {
                if (degreeMap.containsKey(i)) {
                    actorList.addAll(degreeMap.get(i));
                }
            }

            System.out.println(actorList + " have " + currCenter + " numbers between " + lowNum + " and " + upNum + "\n");

        }
    }

    /**
     * Helper class to store average separation of actor when looking for actors with highest / lowest avg. separation
     */
    public class SepAndActor {
        double avgSep;
        String actor;

        public SepAndActor(double avgSep, String actor) {
            this.avgSep = avgSep;
            this.actor = actor;
        }
    }

    public static void main(String[] args) {

        BaconGameInterface tester = new BaconGameInterface(
                "/Users/tuckersimpson/IdeaProjects/cs10/ps4_inputs/actorsTest.txt",
                "/Users/tuckersimpson/IdeaProjects/cs10/ps4_inputs/moviesTest.txt",
                "/Users/tuckersimpson/IdeaProjects/cs10/ps4_inputs/movie-actorsTest.txt");

        tester.newGame();

        while(tester.gameRunning) {
            tester.handleInput();
        }
    }



}
