import java.util.*;

/**
 * PS4
 * Authors: Logan Tucker and Tucker Simpson
 */

public class BaconGraphLib {


    /**
     * BFS to find the shortest path tree for a current center of the universe. Return a path tree as a Graph.
     * @param g
     * @param source
     * @param <V>
     * @param <E>
     * @return graph of costar connections stemming from a specified vertex
     */
    public static <V,E> Graph<V,E> bfs(Graph<V,E> g, V source){

        // create storage
        Graph<V, E> backGraph = new AdjacencyMapGraph<>();
        Set<V> visited = new HashSet<>(); // holds vertices that have already been added to the graph
        Queue<V> queue = new LinkedList<>(); // holds vertices that are yet to be inspected

        queue.add(source);
        visited.add(source);
        backGraph.insertVertex(source);

        // while there are still vertices to be inspected
        while(!queue.isEmpty()){
            V v = queue.remove();

            // for each neighbor of the current vertex (v)
            for(V n : g.outNeighbors(v)){

                // if we have not inspected this vertex before
                if(!visited.contains(n)){

                    // create a connection between current vertex and its neighbor
                    backGraph.insertVertex(n);
                    backGraph.insertDirected(n, v, g.getLabel(v, n));

                    // mark the neighbor as visited to avoid creating connections to it pointing in the wrong direction
                    visited.add(n);

                    // add the neighbor to the que to inspect its neighbors
                    queue.add(n);
                }
            }
        }
        return backGraph;
    }


    /**
     * Given the shortest path tree and a vertex, construct a path from the vertex back to the center of the universe.
     * @param tree
     * @param v
     * @param <V>
     * @param <E>
     * @return return path from vertex to center of the graph
     */
    public static <V,E> List<V> getPath(Graph<V,E> tree, V v){

        if(!tree.hasVertex(v)) return null;

        // create storage
        ArrayList<V> path = new ArrayList<>();
        V curr = v;

        // while the current is not the center of the graph
        while(tree.outDegree(curr)>0){

            path.add(curr);

            // set current to a neighbor (each vertex only has 1 out neighbor)
            for(V neighbor: tree.outNeighbors(curr)) {
                curr = neighbor;
            }
        }

        // finally, add the center of the graph to the path
        path.add(curr);

        return path;
    }


    /**
     * Given a graph and a subgraph (here the shortest path tree), determine which vertices are in the graph but not
     * the subgraph (here, not reached by BFS).
     * @param graph
     * @param subgraph
     * @param <V>
     * @param <E>
     * @return a list of values with an infinite degree of separation from the center of the graph
     */
    public static <V,E> Set<V> missingVertices(Graph<V,E> graph, Graph<V,E> subgraph){

        Set<V> missing = new HashSet<>();

        // for every vertex in the greater graph
        for(V v : graph.vertices()){
            if(!(subgraph.hasVertex(v))) {
                missing.add(v);
            }
        }
        return missing;
    }


    /**
     * Find the average distance-from-root in the shortest path tree.
     * @param tree
     * @param root
     * @param <V>
     * @param <E>
     * @return (double) average distance-from-root in the shortest path tree
     */
    public static <V,E> double averageSeparation(Graph<V,E> tree, V root) {

        double numberOfVertices = tree.numVertices()-1; // subtract one to account for the center itself
        if (numberOfVertices == 0) return 0;
        double sumOfAll = totalLength(tree, root, 0);

        return sumOfAll / numberOfVertices;
    }


    /**
     * Helper method for averageSeparation
     * This method returns the sum of every vertex's degree of separation from the center of a graph
     * @param tree
     * @param vertex
     * @param currDegreeSep
     * @param <V>
     * @param <E>
     * @return sum of every vertex's degree of separation from the center
     */
    public static <V,E> double totalLength(Graph<V,E> tree, V vertex, int currDegreeSep){

        double total = currDegreeSep;

        // for each neighbor of the current vertex
        for(V v : tree.inNeighbors(vertex)){

            // recursively call totalLength with an added degree of separation.
            total += totalLength(tree, v, currDegreeSep+1);
        }
        return total;
    }


    /**
     * This method sorts vertices in decreasing order by their in-degree
     * @param g
     * @return list of vertices sorted by in-degree, decreasing (i.e., largest at index 0)
     */
    public static <V,E> List<V> verticesByInDegree(Graph<V,E> g) {

        PriorityQueue<V> v = new PriorityQueue<>((V v1, V v2) -> ((g.inDegree(v2) - g.outDegree(v1))));

        for(V vertex : g.vertices()) {
            v.add(vertex);
        }

        ArrayList<V> actorsByInDegree = new ArrayList<>();
        int size = v.size();

        for (int i = 0; i < size; i++) {
            actorsByInDegree.add(v.remove());
        }

        return actorsByInDegree;
    }
}
